#!/usr/bin/perl

# Program:	tipsterize_tdt.perl
# Written by:	Dave Graff, LDC
# Purpose:	read token stream and boundary table files to create 
#		corresponding "TIPSTER" style SGML files in a TDT text corpus

$Usage = "Usage: $0 -i input_type -o output_dir [-t table_dir] [-nonempty]\n";

# In TDT evaluations, the systems being tested are supposed to use
# "token stream" files as input.  In token streams, each line of data
# contains a single word with a sequential ID number; each word is
# marked with a "<W recid=N>" SGML tag.  In the case of ASR files,
# there are also lines of data that mark pauses between some words
# (using a "<X ...>" tag), and extra attributes to indicate time
# offsets for the word or pause (relative to the start of the audio
# data file), plus any available information about speaker-cluster
# identification and the recognition confidence score for each word.
# For example:

# <DOCSET type=ASRTEXT fileid=19981001_2100_2200_MNB_NBW ...>
# <X Bsec=0.00 Dur=2.49 Conf=NA>
# <W recid=1 Bsec=2.49 Dur=0.31 Clust=NA Conf=NA> A
# <W recid=2 Bsec=2.83 Dur=0.16 Clust=NA Conf=NA> VERY
# <W recid=3 Bsec=2.99 Dur=0.34 Clust=NA Conf=NA> DOWN
# <W recid=4 Bsec=3.33 Dur=0.26 Clust=NA Conf=NA> DAY
# ...
# <W recid=9201 Bsec=3657.26 Dur=0.08 Clust=NA Conf=NA> THE
# <W recid=9202 Bsec=3657.34 Dur=0.46 Clust=NA Conf=NA> MOST
# <X Bsec=3657.80 Dur=1.76 Conf=NA>
# <W recid=9203 Bsec=3659.56 Dur=0.38 Clust=NA Conf=NA> NO
# </DOCSET>

# Story boundaries, DOCNO identifiers for stories, and categorization
# of story content as "news" or "misc" are not provided at all in the
# token stream file; instead, this information is found in the
# corresponding boundary table file, which contains one entry for each
# story unit in the file.  The attributes in each table entry indicate
# the starting and ending word ids (if the unit contains any words),
# the starting and ending time offsets (if its a broadcast file), the
# DOCNO, and the category for the corresponding story unit; for
# example:

# <BOUNDSET type=ASRTEXT fileid=19981001_2100_2200_MNB_NBW ...
# <BOUNDARY docno=MNB19981001.2100.0000 doctype=MISCELLANEOUS Bsec=0.00 Esec=67.57 Brecid=1 Erecid=164>
# <BOUNDARY docno=MNB19981001.2100.0067 doctype=NEWS Bsec=67.57 Esec=265.73 Brecid=165 Erecid=826>
# ...
# <BOUNDARY docno=MNB19981001.2100.3409 doctype=NEWS Bsec=3409.81 Esec=3563.66 Brecid=8586 Erecid=9015>
# <BOUNDARY docno=MNB19981001.2100.3563 doctype=MISCELLANEOUS Bsec=3563.66 Esec=3659.94 Brecid=9016 Erecid=9203>
# </BOUNDSET>

# This perl script is intended to create a more accessible file format
# for the token-stream data.  In particular, it creates a "TIPSTER"
# style SGML version of each file, using markup that is comparable to
# the "reference-text" SGML format; for example:

# <DOC>
# <DOCNO> MNB19981001.2100.0000 </DOCNO>
# <DOCTYPE> MISCELLANEOUS </DOCTYPE>
# <TXTTYPE> ASRTEXT </TXTYPE>
# <TEXT>
# A VERY DOWN DAY ...
# </TEXT>
# </DOC>

############## NOTABLE DETAIL: ################
# In broadcast data, some "story units" contain no text (no words) at
# all (and the units that lack text may vary with the type of data --
# i.e. the ASR file might contain words for a unit while the TKN file
# might not, or vice-versa).  By default, "tipsterize_tdt.perl" will
# output all story units, whether or not they contain text; units with
# no text will be rendered with all the TIPSTER SGML markup outlined
# above, and there will be nothing between <TEXT> and </TEXT>.
#
# If you would prefer that empty story units be excluded from the
# output, use the optional "-nonempty" argument on the command line.
###############################################

# The script must be run from the base directory of the TDT corpus;
# the command line arguments must specify the type of token stream to
# be processed (e.g. "-i as0" or "-i mttkn"), and a directory path
# where output files will be written (e.g. "-o as0_sgm"); by default,
# the "ground truth" boundary tables for the chosen data type will be
# used to create the output format, but an alternative set of boundary
# tables may be specified (using e.g. "-t as0_othr_bnd").  Command
# line args can be given in any order (so long as each path name comes
# directly after its respective option flag).

$keepEmpties = 1;
( @ARGV >= 4 ) || die $Usage;

while ( @ARGV ) {
    $opt = shift;
    if ( $opt eq "-nonempty" ) {
	$keepEmpties = 0;
	next;
    }
    $arg = shift;
    if ( $opt eq "-i" ) {
	$typ = $arg;
    } elsif ( $opt eq "-o" ) {
	$out = $arg;
    } elsif ( $opt eq "-t" ) {
	$typ_bnd = $arg;
    }
}
$typ_bnd = $typ . "_bnd" if ( $typ_bnd eq "" );

( $typ && $out && -d $typ && -d $typ_bnd ) || die $Usage;
unless ( -d $out ) {
    mkdir $out, 0755 || die "Unable to create output directory ./$out\n$Usage";
}

opendir( DIR, $typ_bnd ) || die "Unable to opendir $typ_bnd: $!\n";
@tblfiles = sort( grep { /\.$typ_bnd$/ } readdir( DIR ));
closedir DIR;

opendir( DIR, $typ ) || die "Unable to opendir $typ: $!\n";
@strfiles = sort( grep { /\.$typ$/ } readdir( DIR ));
closedir DIR;

(scalar @tblfiles) || die "Odd, there are no *.$typ_bnd files in the $typ_bnd directory.\n";
(scalar @strfiles) || die "Odd, there are no *.$typ files in the $typ directory.\n";

for $t (@tblfiles)
{
    $base = $t;
    $base =~ s/$typ_bnd//;
    ($f) = grep /$base/, @strfiles;
    if ( $f eq "" ) {
	warn "No token stream found for $typ_bnd/$t; skipping this table file\n";
	next;
    }
    unless ( open( INP, "$typ_bnd/$t" )) {
	warn "Can't open boundary table $typ_bnd/$t: $!  (skipping it)\n";
	next;
    }
    @storylist = <INP>;
    close INP;

    unless ( open( INP, "$typ/$f" )) {
	warn "Can't open token stream $typ/$f: $!  (skipping it)\n";
	next;
    }
    @tknlist = <INP>;
    close INP;

    %bgn = ();
    %end = ();
    %dtp = ();
    @dns = ();
    $lastrec = 0;

    for ( @storylist )
    {
	if ( /BOUNDSET type=(\S+)/o ) {
	    $ftyp = $1;
	    $pad = ( /src_lang=MANDARIN content_lang=NATIVE/  # manually-created Mandarin
		     && /(NEWSWIRE|TRANSCRIPT)/ ) ?           # text does not have spaces
			  "" : " ";            # between "words"; all other data types do
	    next;
	}
	if ( /docno=(\S+) doctype=(\S+)/o ) {
	    $dn = $1;
	    $dtp{$dn} = $2;
	    push( @dns, $dn );
	}
	if ( /Brecid=(\d+) Erecid=(\d+)/o ) {
	    $bgn{$dn} = $1;
	    $end{$dn} = $2;
	    if ( $bgn{$dn} <= $lastrec || $end{$dn} < $bgn{$dn} ) {
		warn "Bad boundary for $dn in $t: bgn=$bgn{$dn} end=$end{$dn} (treating as empty)\n";
		delete $bgn{$dn};
		delete $end{$dn};
	    } else {
		$lastrec = $end{$dn};
	    }
	}
    }
    $_ = "";
    $fo = $f;
    $fo =~ s/$typ$/$out/;  # make the output file extension match the output path name
    open( OUT, ">$out/$fo" ) || die "Unable to open output file $out/$fo: $!\n";
    for $dn ( @dns )
    {
	next unless ( $keepEmpties || exists( $bgn{$dn} ));
	print OUT
	    "<DOC>\n<DOCNO> $dn </DOCNO>\n<DOCTYPE> $dtp{$dn} </DOCTYPE>\n".
	    "<TXTTYPE> $ftyp </TXTTYPE>\n<TEXT>\n";
	$ccount = 0;
	if ( exists( $bgn{$dn} )) {
	    until ( /recid=$bgn{$dn}/ || @tknlist == 0 ) {
		$_ = shift( @tknlist );
	    }
	    $line = "";
	    $llen = 0;
	    while (@tknlist) {
		if ( /^<W /o ) {
		    @e = split;
		    $wrd = pop( @e );
		    if ( $pad ) {
			$wrd .= $pad;
		    } elsif ( $wrd =~ /[\!-\~]/o ) { # ascii content within manual GB 
			$wrd = " $wrd ";             # text should be space-bounded
		    }
		    $line .= $wrd;
		    $llen += length( $wrd );
		    if ( $llen > 65 ) {
			print OUT "$line\n";
			$llen = 0;
			$line = "";
		    }
		    if ( /recid=$end{$dn}/ ) {
			print OUT "$line\n" if ( $line ne "" );
			last;
		    }
		}
		$_ = shift( @tknlist );
	    }
	}
	print OUT "</TEXT>\n</DOC>\n";
    }
    close OUT;
}
